<?php

namespace WooCommerce\Connectix;

use Connectix\Components\Integration;
use Connectix\Components\IntegrationHook;
use Connectix\Components\IntegrationHookUnregister;
use Connectix\Components\IntegrationUnregister;
use Connectix\Exception;
use Connectix\Request\BaseRequest;
use WP_Error;
use WP_HTTP_Response;
use WP_REST_Request;
use WP_REST_Response;

defined('ABSPATH') || exit;

/**
 * Class Api
 * @package WooCommerce\Connectix
 */
class Api extends BaseClient
{
    /**
     * Add routes
     */
    public function add_routes()
    {
        register_rest_route(
            'connectix-api',
            '/settings',
            [
                'methods' => 'POST',
                'callback' => [$this, 'handler'],
                'args' => [
                    'ajax' => [
                        'type' => 'integer',
                        'required' => true,
                        'sanitize_callback' => 'absint',
                    ],
                    'action' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'enabled' => [
                        'type' => 'boolean',
                        'required' => false,
                    ],
                    'system' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'token' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'addContacts' => [
                        'type' => 'boolean',
                        'required' => false,
                    ],
                    'trackDelivery' => [
                        'type' => 'boolean',
                        'required' => false,
                    ],
                    'deliveredStatus' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'sandboxUrl' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'connectixToken' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'confirmOrder' => [
                        'type' => 'boolean',
                        'required' => false,
                    ],
                    'confirmedOrderStatus' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'confirmedOrderUrl' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'cancelOrder' => [
                        'type' => 'boolean',
                        'required' => false,
                    ],
                    'canceledOrderStatus' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'canceledOrderUrl' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                    'confirmationUrlsValidity' => [
                        'type' => 'integer',
                        'required' => false,
                        'sanitize_callback' => 'absint',
                    ],
                    'confirmationUrlsExpiredUrl' => [
                        'type' => 'string',
                        'required' => false,
                        'sanitize_callback' => 'sanitize_text_field',
                    ],
                ],
                'permission_callback' => [$this, 'permissions'],
            ]
        );
    }

    /**
     * @deprecated to be removed if permission_callback works
     * Check request permissions
     *
     * @return bool
     */
    public function permissions(): bool
    {
        return current_user_can('edit_others_shop_orders');
    }

    /**
     * Update settings
     *
     * @param WP_REST_Request $request
     * @return WP_Error|WP_HTTP_Response|WP_REST_Response
     */
    public function handler(WP_REST_Request $request)
    {
        switch ($request->get_param('action')) {
            case 'getData':
                $data = [
                    'settings' => $this->getSettings(),
                    'orderStatuses' => $this->fetchOrderStatuses(),
                    'templates' => $this->fetchTemplates(),
                    'rules' => $this->fetchRules(),
                ];
                break;
            case 'saveSettings':
                $postData = [];
                $oldSettings = $this->getSettings(true);
                foreach (array_keys($oldSettings) as $setting) {
                    $name = $setting;
                    if ('token' === $setting) {
                        $name = 'connectixToken';
                    }

                    if ($request->has_param($name)) {
                        $postData[$setting] = $request->get_param($name);
                    } else {
                        $postData[$setting] = $oldSettings[$setting];
                    }
                }

                if (!empty($postData)) {
                    Settings::save_settings($postData);
                }

                $newSettings = $this->getSettings(true);

                $isTokenChanged = $oldSettings['token'] !== $newSettings['token'];
                if ($isTokenChanged && $oldSettings['token'] && $oldSettings['integrationToken']) {
                    $data = new IntegrationUnregister();
                    $data->setSecret($oldSettings['secret'])
                        ->setToken($oldSettings['integrationToken']);
                    try {
                        $this->getClient()->setToken($oldSettings['token'])->getIntegration()->unregister($data);
                    } catch (\Exception $e) {}
                    $postData['integrationToken'] = null;
                }

                $templates = null;
                if ($isTokenChanged && $newSettings['token']) {
                    $data = new Integration();
                    $data->setSecret($newSettings['secret'])
                        ->setPlatform('WooCommerce')
                        ->setType(Integration::TYPE_ECOMMERCE);
                    try {
                        $response = $this->getClient()->setToken($newSettings['token'])->getIntegration()->register($data);
                    } catch (\Exception $e) {
                        $response = null;
                    }

                    if ($response) {
                        $postData['integrationToken'] = $response['token'];
                        $templates = $this->syncTemplates();
                    }
                }

                if (!empty($postData)) {
                    Settings::save_settings($postData);
                    $newSettings = $this->getSettings(true);
                }
                $this->unregisterHooks($oldSettings, $newSettings);
                $this->registerHooks($oldSettings, $newSettings);

                $data = $this->getSettings();
                if ($templates) {
                    $data['templates'] = $templates;
                }
                break;
            case 'syncTemplates':
                $data = $this->syncTemplates();
                break;
            case 'saveRule':
                $data = $this->saveRule($request);
                break;
            case 'deleteRule':
                $data = $this->deleteRule($request);
                break;
            default:
                $data = [];
                break;
        }

        return rest_ensure_response($data);
    }

    /**
     * @param array $oldSettings
     * @param array $newSettings
     */
    private function unregisterHooks(array $oldSettings, array $newSettings)
    {
        $shouldUnregister = $oldSettings['token'] !== $newSettings['token'] || ($oldSettings['enabled'] !== $newSettings['enabled'] && !$newSettings['enabled']);

        $postData = [];
        $client = $this->getClient()->setToken($oldSettings['token']);
        foreach ([IntegrationHook::TYPE_CALLBACK, IntegrationHook::TYPE_INBOUND, IntegrationHook::TYPE_TEMPLATE, IntegrationHook::TYPE_TRACKING] as $type) {
            $shouldUnregisterTrackingHook = $type === IntegrationHook::TYPE_TRACKING && $oldSettings['trackDelivery'] !== $newSettings['trackDelivery'] && !$newSettings['trackDelivery'];
            if ($shouldUnregister || $shouldUnregisterTrackingHook) {
                $hookSettingsName = 'hook'.ucfirst($type);
                $hookId = $newSettings[$hookSettingsName];
                if (!empty($hookId) && $oldSettings['token']) {
                    $data = new IntegrationHookUnregister();
                    $data->setId($hookId);
                    try {
                        $client->getIntegrationHook()->unregister($data);
                    } catch (\Exception $e) {}
                    $postData[$hookSettingsName] = null;
                }
            }
        }

        if (!empty($postData)) {
            Settings::save_settings($postData);
        }
    }

    /**
     * @param array $oldSettings
     * @param array $newSettings
     */
    private function registerHooks(array $oldSettings, array $newSettings)
    {
        // Hooks callback|inbound|template|tracking
        $token = $newSettings['token'];
        $secret = $newSettings['secret'];
        $integrationToken = $newSettings['integrationToken'];

        $shouldRegister = $integrationToken && (($oldSettings['enabled'] !== $newSettings['enabled'] && $newSettings['enabled']) || ($oldSettings['token'] !== $token && $token && $newSettings['enabled']));

        $postData = [];
        $client = $this->getClient()->setToken($token);
        foreach ([IntegrationHook::TYPE_CALLBACK, IntegrationHook::TYPE_INBOUND, IntegrationHook::TYPE_TEMPLATE, IntegrationHook::TYPE_TRACKING] as $type) {
            $shouldRegisterTrackingHook = $integrationToken && $newSettings['enabled'] && $type === IntegrationHook::TYPE_TRACKING && $oldSettings['trackDelivery'] !== $newSettings['trackDelivery'] && $newSettings['trackDelivery'];
            if (($type !== IntegrationHook::TYPE_TRACKING && $shouldRegister) || $shouldRegisterTrackingHook) {
                $hookSettingsName = 'hook'.ucfirst($type);
                $url = add_query_arg(['type' => $type, 'secret' => $secret], rest_url('connectix-api/callback'));
                $data = new IntegrationHook();
                $data->setType($type)
                    ->setToken($integrationToken)
                    ->setUrl($url);

                try {
                    $response = $client->getIntegrationHook()->register($data);
                    $postData[$hookSettingsName] = $response['id'];
                } catch (\Exception $e) {}
            }
        }

        if (!empty($postData)) {
            Settings::save_settings($postData);
        }
    }

    /**
     * @param $path
     * @param $token
     * @param array $data
     * @param array $params
     * @return false|mixed
     */
    public static function call($path, $token, array $data = [], array $params = [])
    {
        if (empty($token)) {
            return false;
        }

        $headers = [
            "Content-Type: application/json",
            "Authorization: {$token}",
        ];

        $url = BaseRequest::SANDBOX_ENDPOINT;
        if (Settings::get_setting('system') === self::SYSTEM_LIVE) {
            $url = BaseRequest::LIVE_ENDPOINT;
        }
        $url .= $path;
        if (!empty($params)) {
            $url .= '?'.http_build_query($params);
        }

        $ch = curl_init();

        curl_setopt_array($ch, [
            CURLOPT_URL => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_FOLLOWLOCATION => false,
            CURLOPT_HTTPHEADER => $headers,
            CURLOPT_TIMEOUT => 3,
            CURLOPT_POST => true,
            CURLOPT_POSTFIELDS => json_encode($data),
        ]);

        $error = false;
        if (!$result = curl_exec($ch)) {
            $error = true;
        }

        if (200 !== curl_getinfo($ch, CURLINFO_RESPONSE_CODE) || 0 !== curl_errno($ch)) {
            $error = true;
        }

        curl_close($ch);

        if ($error) {
            return false;
        }

        return json_decode($result, true);
    }

    /**
     * @param bool $full
     * @return array
     */
    public function getSettings($full = false): array
    {
        $settings = Settings::get_settings();

        if (!$full) {
            unset($settings['secret']);
            unset($settings['integrationToken']);
            unset($settings['hookCallback']);
            unset($settings['hookInbound']);
            unset($settings['hookTemplate']);
            unset($settings['hookTracking']);
        }

        return $settings;
    }

    /**
     * @return array
     */
    private function fetchOrderStatuses(): array
    {
        $statuses = [];
        foreach (wc_get_order_statuses() as $status => $name) {
            $statuses[] = [
                'key' => $status,
                'value' => $status,
                'text' => $name,
            ];
        }

        return $statuses;
    }

    /**
     * @return array
     */
    private function syncTemplates(): array
    {
        global $wpdb;

        try {
            $templates = $this->getClient()->getTemplates()->fetchTemplates();
        } catch (Exception\BadConnectionException $e) {
            return [];
        } catch (Exception\BadRequestException $e) {
            return [];
        } catch (Exception\BadResponseBodyException $e) {
            return [];
        }

        $existingTemplates = $this->fetchTemplates();

        $templateIds = array_map(
            function ($template) {
                return $template['id'];
            },
            $templates
        );

        $fields = ['media', 'layout', 'name', 'channel', 'text', 'button_label', 'button_url', 'status'];
        $formatData = function ($templateData) use ($fields, $wpdb) {
            $data = [];
            foreach ($fields as $field) {
                $data[$field] = isset($templateData[$field]) ? $wpdb->_escape($templateData[$field]) : null;
            }

            return $data;
        };
        $isChanged = function ($old, $new) use ($fields) {
            $changed = false;
            foreach ($fields as $field) {
                if (!isset($new[$field])) {
                    continue;
                }

                if ($old[$field] !== (string)$new[$field]) {
                    $changed = true;
                    break;
                }
            }

            return $changed;
        };

        $forDeletion = $forUpdate = [];
        foreach ($existingTemplates as $template) {
            if (!in_array($template['provider_id'], $templateIds)) {
                $forDeletion[] = $template['id'];
            } elseif (false !== $index = array_search($template['provider_id'], array_column($templates, 'id'))) {
                if ($isChanged($template, $templates[$index])) {
                    $forUpdate[$template['id']] = $formatData($templates[$index]);
                }
            }
        }

        $fetchFresh = false;
        // Handle removals
        if (!empty($forDeletion)) {
            $fetchFresh = true;
            $wpdb->query('DELETE FROM '.$wpdb->prefix.'connectix_templates WHERE id IN ('.$wpdb->_escape(join(',', $forDeletion)).')');
        }

        // Handle updates
        foreach ($forUpdate as $id => $data) {
            $fetchFresh = true;
            $wpdb->update($wpdb->prefix.'connectix_templates', $data, ['id' => $id]);
        }

        // Handle insertions
        foreach ($templates as $template) {
            if (false === array_search($template['id'], array_column($existingTemplates, 'provider_id'))) {
                $fetchFresh = true;
                $data = $formatData($template);
                $data['provider_id'] = $template['id'];

                // ToDo: rework this and make it safer
                $insertSet = [];
                foreach ($data as $key => $value) {
                    $insertSet['columns'][] = "`$key`";
                    $insertSet['values'][] = "'$value'";
                }
                $wpdb->query('INSERT INTO ' . $wpdb->prefix . 'connectix_templates (' . implode(',', $insertSet['columns']) . ') VALUES (' . implode(',', $insertSet['values']) . ')');
            }
        }

        // Fetch fresh data if it's necessary
        return $fetchFresh ? $this->fetchTemplates() : $existingTemplates;
    }

    /**
     * @return array
     */
    private function fetchTemplates(): array
    {
        global $wpdb;

        return $wpdb->get_results("SELECT * FROM {$wpdb->prefix}connectix_templates", 'ARRAY_A');
    }

    /**
     * @param string|null $type
     * @param string|int|null $id
     * @return array
     */
    public static function fetchRules($type = null, $id = null): array
    {
        global $wpdb;

        $where = $params = [];
        if ($type && in_array(
                $type,
                [self::TYPE_NEW_ORDER, self::TYPE_CHANGE_ORDER_STATUS, self::TYPE_READY_FOR_DELIVERY, self::TYPE_NOT_PICKED_UP]
            )) {
            $where[] = 'r.type = "%s"';
            $params[] = $type;
        }
        if ($id) {
            $where[] = 'r.id = "%d"';
            $params[] = $id;
        }
        if (!empty($where) && !$id) {
            $where[] = 'r.enabled = "1"';
        }

        $query = $wpdb->prepare('SELECT
                r.*,
                rt.position as rt_position,
                t.id as t_id,
                t.name as t_name,
                t.channel as t_channel,
                t.status as t_status,
                t.provider_id as t_provider_id
            FROM
                '.$wpdb->prefix.'connectix_rules r
            LEFT JOIN '.$wpdb->prefix.'connectix_rules_templates rt ON rt.rule_id = r.id
            LEFT JOIN '.$wpdb->prefix.'connectix_templates t ON rt.template_id = t.id
            '.(!empty($where) ? 'WHERE '.join(' AND ', $where) : '').'
            ORDER BY rt.position ASC', $params);

        $rules = $wpdb->get_results($query, ARRAY_A);
        if (!$rules) {
            return [];
        }

        $normalized = [];
        foreach ($rules as $rule) {
            $template = [
                'id' => $rule['t_id'],
                'position' => $rule['rt_position'],
                'name' => $rule['t_name'],
                'channel' => $rule['t_channel'],
                'status' => $rule['t_status'],
                'provider_id' => $rule['t_provider_id'],
            ];

            if (!isset($normalized[$rule['id']])) {
                $normalized[$rule['id']] = [
                    'id' => $rule['id'],
                    'fallback' => $rule['fallback'] === '1',
                    'type' => $rule['type'] !== '' ? $rule['type'] : null,
                    'order_status' => $rule['order_status'] !== '' ? $rule['order_status'] : null,
                    'delivery_type' => $rule['delivery_type'] !== '' ? $rule['delivery_type'] : null,
                    'days_not_picked_up' => $rule['days_not_picked_up'] !== '' ? $rule['days_not_picked_up'] : null,
                    'enabled' => '1' === $rule['enabled'],
                    'templates' => [$template],
                ];
            } else {
                $normalized[$rule['id']]['templates'][] = $template;
            }
        }

        return array_values($normalized);
    }

    /**
     * @param WP_REST_Request $request
     * @return array|mixed|null
     */
    private function saveRule(WP_REST_Request $request): ?array
    {
        global $wpdb;

        $data = [];
        foreach (['enabled', 'fallback', 'type', 'delivery_type', 'order_status', 'days_not_picked_up'] as $field) {
            $value = $wpdb->_escape($request->get_param($field));
            switch ($value) {
                case 'null':
                    $value = null;
                    break;
                case 'false':
                case 'true':
                    $value = $value === 'true' ? 1 : 0;
                    break;
            }
            $data[$field] = $value;
        }

        if ($request->has_param('id')) {
            $id = (int)$request->get_param('id');
            $wpdb->delete($wpdb->prefix.'connectix_rules_templates', ['rule_id' => $id]);
            $wpdb->update($wpdb->prefix.'connectix_rules', $data, ['id' => $id]);
        } else {
            $wpdb->insert($wpdb->prefix.'connectix_rules', $data);
            $id = $wpdb->insert_id;
        }

        foreach ($request->get_param('templates') as $i => $templateId) {
            $wpdb->insert(
                $wpdb->prefix.'connectix_rules_templates',
                ['template_id' => $templateId, 'rule_id' => $id, 'position' => $i]
            );
        }

        $record = $this->fetchRules(null, $id);
        if (!empty($record)) {
            $record = $record[0];
        }

        return $record;
    }

    /**
     * @param WP_REST_Request $request
     * @return string
     */
    private function deleteRule(WP_REST_Request $request): string
    {
        global $wpdb;

        if ($request->has_param('id')) {
            $id = (int)$request->get_param('id');
            $wpdb->delete($wpdb->prefix.'connectix_rules', ['id' => $id]);

            return "OK";
        }

        // ToDo: throw an error here
        return "Not found";
    }

    ##########################################################
    #                         Helpers                        #
    ##########################################################
    /**
     * @param string $value
     * @return string|null
     */
    private function nullableValue(string $value): ?string
    {
        if ('' === $value) {
            return null;
        }

        return $value;
    }
}
