<?php

namespace WooCommerce\Connectix;

use DateTime;
use WP_REST_Request;

defined('ABSPATH') || exit;

/**
 * Class Callback
 * @package WooCommerce\Connectix
 */
class Callback
{
    /**
     * Add routes
     */
    public function add_routes()
    {
        register_rest_route(
            'connectix-api',
            '/callback',
            [
                'methods' => ['GET', 'POST'],
                'callback' => [$this, 'callback'],
                'permission_callback' => '__return_true',
            ]
        );
    }

    public function callback(WP_REST_Request $request)
    {
        $paramSecret = $request->get_param('secret');
        $secret = Settings::get_setting('secret');
        $response = rest_ensure_response('');

        $type = $request->get_param('type');
        if (!$type) {
            return $response;
        }

        $isSpecialRoute = in_array($type, ['confirm', 'cancel']);
        if (!Settings::get_setting('enabled') || (!$isSpecialRoute && (empty($paramSecret) || $paramSecret !== $secret)) || ($isSpecialRoute && empty($request->get_param('token')))) {
            $response->set_status(401);
            return $response;
        }

        $data = $request->get_json_params();

        if (!$isSpecialRoute && is_null($data)) {
            $response->set_status(400);
            return $response;
        }

        $action = null;
        if (isset($data['action'])) {
            $action = $data['action'];
            unset($data['action']);
        }

        $success = null;
        switch ($type) {
            case 'callback':
                $success = $this->handleCallback($data);
                break;
            case 'inbound':
                $success = $this->handleInbound($data);
                break;
            case 'template':
                $success = $this->handleTemplate($data);
                break;
            case 'tracking':
                if ($action === 'readyForDelivery') {
                    $success = $this->handleReadyForDelivery($data);
                } elseif ($action === 'delivered') {
                    $success = $this->handleDelivered($data);
                } elseif ($action === 'notPickedUp') {
                    $success = $this->handleNotPickedUp($data);
                } else {
                    $success = true;
                }
                break;
            case 'confirm':
            case 'cancel':
                $success = $this->handleOrderConfirmation($type, $request->get_param('token'));
                break;
        }

        if (true === $success) {
            $response->set_data('OK');
        } elseif (null === $success) {
            $response->set_status(202);
            $response->set_data('OK');
        } else {
            $response->set_status(400);
            $response->set_data('Something is wrong..');
        }

        return $response;
    }

    /**
     * @param $expected
     * @param $data
     * @return bool
     */
    private function validateData($expected, $data): bool
    {
        $valid = true;
        foreach ($expected as $field) {
            if (!array_key_exists($field, $data)) {
                $valid = false;
                break;
            }
        }
        return $valid;
    }

    /**
     * @param array $data
     * @return false
     */
    private function handleCallback(array $data): bool
    {
        global $wpdb;

        if (!$this->validateData(['id', 'status', 'channel', 'createdAt', 'fallbackId'],  $data) && !$this->validateData(['id', 'status', 'channel', 'createdAt'],  $data)) {
            return false;
        }

        $message = $wpdb->get_row('SELECT id,fallback_id FROM '.$wpdb->prefix.'connectix_messages WHERE provider_id = "'.$wpdb->_escape($data['id']).'"', ARRAY_A);
        if (!$message) {
            return true;
        }

        $wpdb->update(
            $wpdb->prefix.'connectix_messages',
            [
                'channel' => $data['channel'],
                'status' => $data['status'],
            ],
            ['provider_id' => $data['id']]
        );

        if (in_array($data['status'], ['delivered', 'seen']) && $message['fallback_id']) {
            $wpdb->delete($wpdb->prefix.'connectix_messages', ['status' => 'pending', 'fallback_id' => $message['fallback_id']]);
        }

        return true;
    }

    /**
     * @param array $data
     * @return false
     */
    private function handleInbound(array $data): bool
    {
        global $wpdb;

        if (!$this->validateData(['id', 'type', 'phone', 'text', 'sentAt', 'mediaUrl', 'mediaName'],  $data)) {
            return false;
        }

        $message = $wpdb->get_row('SELECT id FROM '.$wpdb->prefix.'connectix_messages WHERE provider_id = "'.$wpdb->_escape($data['id']).'"', ARRAY_A);
        if (!$message) {
            return true;
        }

        $values = [
            'message_id' => (int)$message['id'],
            'type' => $data['type'],
            'sent_at' => $data['sentAt'] ?? null,
        ];
        if (!empty($data['text'])) {
            $values['text'] = $data['text'];
        }
        if (!empty($data['mediaUrl']) && !empty($data['mediaName'])) {
            $values['media_url'] = $data['mediaUrl'];
            $values['media_name'] = $data['mediaName'];
        }

        $wpdb->insert($wpdb->prefix.'connectix_message_inbounds', $values);

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleTemplate(array $data): bool
    {
        global $wpdb;

        if (!$this->validateData(['id', 'media', 'layout', 'name', 'channel', 'text', 'button_label', 'button_url', 'status'],  $data)) {
            return false;
        }

        $template = $wpdb->get_row(
            'SELECT id FROM '.$wpdb->prefix.'connectix_templates WHERE provider_id = "'.$wpdb->_escape($data['id']).'"',
            ARRAY_A
        );

        $fields = ['media', 'layout', 'name', 'channel', 'text', 'button_label', 'button_url', 'status'];
        $fieldData = [];
        foreach ($fields as $field) {
            $fieldData[$field] = $data[$field];
        }

        if ($template) {
            $wpdb->update(
                $wpdb->prefix.'connectix_templates',
                $fieldData,
                ['provider_id' => $data['id']]
            );
        } else {
            $fieldData['provider_id'] = $data['id'];
            $wpdb->insert($wpdb->prefix.'connectix_templates', $fieldData);
        }

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleReadyForDelivery(array $data): bool
    {
        global $wpdb;

        $expected = ['trackingNumber', 'deliveryType'];
        if (isset($data['deliveryType']) && $data['deliveryType'] === 'door') {
            $expected[] = 'courier';
        } elseif (isset($data['deliveryType']) && $data['deliveryType'] === 'office') {
            $expected[] = 'office';
        }

        if (!$this->validateData($expected,  $data)) {
            return false;
        }

        $trackingOrders = $wpdb->get_results(
            'SELECT order_id FROM '.$wpdb->prefix.'connectix_tracking WHERE tracking_number = "'.$wpdb->_escape($data['trackingNumber']).'"',
            ARRAY_A
        );

        $tmp = [];
        foreach ($trackingOrders as $trackingOrder) {
            $order = wc_get_order((int)$trackingOrder['order_id']);

            if ($order) {
                $tmp[] = [
                    'order' => $order,
                    'deliveryData' => $data,
                ];
            }
        }

        if (!empty($tmp)) {
            do_action(Events::ACTION_NAME_READY_FOR_DELIVERY, ['orders' => $tmp]);
        }

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleNotPickedUp(array $data): bool
    {
        global $wpdb;

        if (!$this->validateData(['trackingNumber', 'workingDaysPassed', 'office'],  $data)) {
            return false;
        }

        $rules = array_filter(Api::fetchRules(Api::TYPE_NOT_PICKED_UP), function ($rule) use ($data) {
            return (int)$data['workingDaysPassed'] === (int)$rule['days_not_picked_up'];
        });

        if (empty($data['trackingNumber'])) {
            return false;
        }

        if (empty($rules)) {
            return true;
        }

        $trackingOrders = $wpdb->get_results(
            'SELECT order_id FROM '.$wpdb->prefix.'connectix_tracking WHERE tracking_number = "'.$wpdb->_escape($data['trackingNumber']).'"',
            ARRAY_A
        );

        $tmp = [];
        foreach ($trackingOrders as $trackingOrder) {
            $order = wc_get_order((int)$trackingOrder['order_id']);

            if ($order) {
                $tmp[] = [
                    'order' => $order,
                    'deliveryData' => $data,
                ];
            }
        }

        if (!empty($tmp)) {
            do_action(Events::ACTION_NAME_NOT_PICKED_UP, [
                'orders' => $tmp,
                'rules' => $rules,
            ]);
        }

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleDelivered(array $data): bool
    {
        global $wpdb;

        if (empty($data['trackingNumber'])) {
            return false;
        }

        $deliveredStatus = Settings::get_setting('deliveredStatus');
        if (!$deliveredStatus) {
            return true;
        }

        $trackingOrders = $wpdb->get_results(
            'SELECT order_id FROM '.$wpdb->prefix.'connectix_tracking WHERE tracking_number = "'.$wpdb->_escape($data['trackingNumber']).'"',
            ARRAY_A
        );

        foreach ($trackingOrders as $trackingOrder) {
            $order = wc_get_order((int)$trackingOrder['order_id']);

            if ($order && $order->get_status() !== $deliveredStatus) {
                $order->update_status($deliveredStatus);
            }
        }

        return true;
    }

    /**
     * @param $action
     * @param $token
     * @return bool
     */
    private function handleOrderConfirmation($action, $token): bool
    {
        global $wpdb;

        $settings = Settings::get_settings();
        $linksValidity = $settings['confirmationUrlsValidity'];
        $expiredUrl = $settings['confirmationUrlsExpiredUrl'];

        $enabled = $settings[$action.'Order'];
        $orderStatus = $settings[$action.'edOrderStatus'];
        $redirectUrl = $settings[$action.'edOrderUrl'];

        if (empty($redirectUrl)) {
//            $redirectUrl = Context::getContext()->shop->getBaseURL(true); // ToDo: find shop url
        }

        if (!$enabled || !$orderStatus) {
            header("Location: $redirectUrl");
            exit;
        }

        $data = $wpdb->get_row(
            'SELECT order_id, created_at FROM '.$wpdb->prefix.'connectix_order_token WHERE token = "'.$wpdb->_escape($token).'"',
            ARRAY_A
        );

        if ($data) {
            $diff = (time() - strtotime($data['created_at'])) / 3600;
            if (0 < $linksValidity && $linksValidity < $diff) {
                if (!$expiredUrl) {
                    echo json_encode('The link expired');
                }

                header("Location: $expiredUrl");
                exit;
            }

            $order = wc_get_order((int)$data['order_id']);
            if ($order && $order->get_status() !== $orderStatus) {
                $order->update_status($orderStatus);
            }

            $wpdb->update(
                $wpdb->prefix.'connectix_order_token',
                [
                    'used' => 1,
                    'status' => $action.'ed',
                    'updated_at' => (new DateTime('now'))->format('Y-m-d H:i:s'),
                ],
                [ 'token' => $token ]
            );
        }

        header("Location: $redirectUrl");
        exit;
    }
}
