<?php

namespace WooCommerce\Connectix;

use Connectix\Exception;
use Connectix\Components\Blacklist;
use Connectix\Components\Fallback;
use Connectix\Components\Contact;
use Connectix\Components\Message;
use WC_Order;

defined('ABSPATH') || exit;

/**
 * Class Events
 * @package WooCommerce\Connectix
 */
class Events extends BaseClient
{
    const ACTION_NAME_READY_FOR_DELIVERY = 'connectix_ready_for_delivery';
    const ACTION_NAME_NOT_PICKED_UP = 'connectix_not_picked_up';

    public function init()
    {
        add_action('woocommerce_checkout_order_created', [$this, 'new_order']);
        add_action('woocommerce_order_status_changed', [$this, 'change_status'], 11, 4);
        add_action('updated_postmeta', [$this, 'processed_metadata'], 11, 4);
        add_action('added_post_meta', [$this, 'processed_metadata'], 11, 4);
        add_action('econt_add_loading', [$this, 'econt_add_loading']);
        add_action('speedy_add_loading', [$this, 'speedy_add_loading'], 10, 2);

        add_action(self::ACTION_NAME_READY_FOR_DELIVERY, [$this, 'ready_for_delivery']);
        add_action(self::ACTION_NAME_NOT_PICKED_UP, [$this, 'not_picked_up']);
    }

    public function processed_metadata($meta_id, $object_id, $meta_key, $meta_value)
    {
        switch($meta_key) {
            case '_wc_shipment_tracking_items':
                $data = maybe_unserialize($meta_value);
                if (!is_array($data) || empty($data)) {
                    return;
                }
                $latest = end($data);
                $trackingNumber = $latest['tracking_number'];
                break;
            case '_order_waybill_id':
                $trackingNumber = $meta_value;
                break;
            default:
                return;
        }

        $this->processTracking($object_id, $trackingNumber);
    }

    public function econt_add_loading($data)
    {
        $this->processTracking($data['order_id'], $data['loading_num']);
    }

    public function speedy_add_loading($orderId, $data)
    {
        $this->processTracking($orderId, $data['id']);
    }

    /**
     * @param $orderId
     * @param $trackingNumber
     */
    private function processTracking($orderId, $trackingNumber)
    {
        global $wpdb;

        if (!$this->isEnabled()) {
            return;
        }

        $tracking = $wpdb->get_row('SELECT id FROM '.$wpdb->prefix.'connectix_tracking WHERE order_id = "'.$wpdb->_escape($orderId).'" AND tracking_number = "'.$wpdb->_escape($trackingNumber).'"', ARRAY_A);
        if ($tracking) {
            $wpdb->update($wpdb->prefix.'connectix_tracking', [
                'tracking_number' => $trackingNumber,
            ], ['id' => $tracking['id']]);
        } else {
            $wpdb->insert($wpdb->prefix.'connectix_tracking', [
                'order_id' => $orderId,
                'tracking_number' => $trackingNumber,
            ]);
        }

        $data = [
            'trackingNumber' => $trackingNumber,
        ];
        $params = [
            'integration' => Settings::get_setting('integrationToken'),
        ];

        Api::call('trackings', Settings::get_setting('token'), $data, $params);
    }

    /**
     * @return int
     */
    private function isEnabled(): int
    {
        return (int)Settings::get_setting('enabled');
    }

    private function getDefaultParameters(WC_Order $order)
    {
        $shipping_address = trim(str_replace($order->get_formatted_shipping_full_name(), '', str_replace('<br/>', ' ', $order->get_formatted_shipping_address())));

        $firstName = $order->get_shipping_first_name();
        $lastName = $order->get_shipping_last_name();
        if (!$firstName) {
            $firstName = $order->get_billing_first_name();
        }
        if (!$lastName) {
            $lastName = $order->get_billing_last_name();
        }

        $productsCount = 0;
        $currency = $order->get_currency();

        $formattedProducts = [];
        foreach ($order->get_items() as $product) {
            $productName = $product->get_name();
            if (50 < mb_strlen($productName)) {
                $productName = trim(mb_substr($productName, 0, 27)).' ... '.trim(mb_substr($productName, -20));
            }

            $quantity = $product->get_quantity();
            $price = $this->formatPrice($product->get_total(), $currency);

            $productsCount += $quantity;
            $formattedProducts[] = "{$productName} x {$quantity} = {$price}";
        }

        $orderTotals = $order->get_order_item_totals();
        $totals = array_values(array_filter($orderTotals, function ($total) {
            return 'payment_method' !== $total;
        }, ARRAY_FILTER_USE_KEY));
        $totals = array_map(function ($total) {
            return $total['label'].' '.$this->formatPrice($total['value'], true);
        }, $totals);

        return [
            'order_number' => (string)$order->get_order_number(),
            'first_name' => (string)$firstName,
            'last_name' => (string)$lastName,
            'total_amount' => $this->formatPrice($order->get_total(), $currency),
            'products' => join("\n\n", $formattedProducts),
            'totals' => join("\n", $totals),
            'payment_method' => $order->get_payment_method_title(),
            'delivery_type' => $order->get_shipping_method(),
            'delivery_address' => $shipping_address,
            'products_count' => (string)$productsCount,
        ];
    }

    /**
     * @return string
     */
    private function generateOrderToken()
    {
        $characters = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
        $charactersLength = strlen($characters);
        $randomString = '';
        for ($i = 0; $i < 32; $i++) {
            $randomString .= $characters[rand(0, $charactersLength - 1)];
        }
        return $randomString;
    }

    /**
     * @param WC_Order $order
     */
    public function new_order(WC_Order $order)
    {
        global $wpdb;

        if (!$this->isEnabled()) {
            return;
        }

        $token = null;
        do {
            $tmp = $this->generateOrderToken();

            $check = $wpdb->get_row('SELECT COUNT(*) as count FROM '.$wpdb->prefix.'connectix_order_token WHERE token = "'.$wpdb->_escape($tmp).'"', ARRAY_A);
            if ($check && !(int)$check['count']) {
                $token = $tmp;
            }
        } while(is_null($token));

        $now = date('Y-m-d H:i:s', time());
        $wpdb->insert($wpdb->prefix.'connectix_order_token', [
            'order_id' => $order->get_id(),
            'token' => $token,
            'created_at' => $now,
            'updated_at' => $now,
        ]);

        $rules = Api::fetchRules(self::TYPE_NEW_ORDER);
        if (empty($rules)) {
            return;
        }

        $phone = $order->get_shipping_phone();
        if (empty($phone)) {
            $phone = $order->get_billing_phone();
        }

        if ($this->isPhoneBlacklisted($phone)) {
            return;
        }

        $parameters = $this->getDefaultParameters($order);
        $contact = new Contact();
        $contact->setFirstName($parameters['first_name'])
            ->setLastName($parameters['last_name']);

        foreach (['confirm', 'cancel'] as $action) {
            $parameters[$action.'_url'] = '';

            if ((int)Settings::get_setting($action.'Order')) {
                $parameters[$action.'_url'] = esc_url_raw(add_query_arg(["type" => $action, "token" => $token], rest_url('connectix-api/callback')));
            }
        }

        foreach ($rules as $rule) {
            $this->sendMessage($order->get_id(), $phone, $rule, $parameters, $contact);
        }
    }

    /**
     * @param int $orderId
     * @param string $statusFrom
     * @param string $statusTo
     * @param WC_Order $order
     */
    public function change_status(int $orderId, string $statusFrom, string $statusTo, WC_Order $order)
    {
        global $wpdb;

        if (!$this->isEnabled()) {
            return;
        }

        $rules = array_filter(Api::fetchRules(self::TYPE_CHANGE_ORDER_STATUS), function ($rule) use ($statusTo) {
            return $rule['order_status'] === 'wc-'.$statusTo; // ToDo: verify this "WC-" part here
        });

        if (empty($rules)) {
            return;
        }

        $phone = $order->get_billing_phone();
        if (empty($phone)) {
            $phone = $order->get_shipping_phone();
        }

        if ($this->isPhoneBlacklisted($phone)) {
            return;
        }

        $parameters = $this->getDefaultParameters($order);

        $contact = new Contact();
        $contact->setFirstName($parameters['first_name'])
            ->setLastName($parameters['last_name']);

        $trackingData = $wpdb->get_row("SELECT tracking_number FROM `".$wpdb->prefix."connectix_tracking` WHERE order_id = '".$orderId."'", ARRAY_A);
        $parameters['tracking_number'] = !empty($trackingData['tracking_number']) ? $trackingData['tracking_number'] : '';

        foreach ($rules as $rule) {
            $this->sendMessage($orderId, $phone, $rule, $parameters, $contact);
        }
    }

    /**
     * @param array $params
     */
    public function ready_for_delivery(array $params)
    {
        global $wpdb;

        if (!$this->isEnabled()) {
            return;
        }

        $rules = Api::fetchRules(self::TYPE_READY_FOR_DELIVERY);
        if (empty($rules)) {
            return;
        }

        $doorRules = array_filter($rules, function ($rule) {
            return 'door' === $rule['delivery_type'];
        });
        $officeRules = array_filter($rules, function ($rule) {
            return 'office' === $rule['delivery_type'];
        });

        foreach ($params['orders'] as $tmp) {
            /** @var WC_Order $order */
            $order = $tmp['order'];
            $deliveryData = $tmp['deliveryData'];

            $phone = $order->get_billing_phone();

            if ($this->isPhoneBlacklisted($phone)) {
                continue;
            }

            $parameters = $this->getDefaultParameters($order);

            $trackingData = $wpdb->get_row("SELECT tracking_number FROM `".$wpdb->prefix."connectix_tracking` WHERE order_id = '".$order->get_id()."'", ARRAY_A);
            $parameters['tracking_number'] = !empty($trackingData['tracking_number']) ? $trackingData['tracking_number'] : '';

            $contact = new Contact();
            $contact->setFirstName($parameters['first_name'])
                ->setLastName($parameters['last_name']);

            if ('office' === $deliveryData['deliveryType']) {
                $parameters['office'] = $deliveryData['office'];
                foreach ($officeRules as $rule) {
                    $this->sendMessage($order->get_id(), $phone, $rule, $parameters, $contact);
                }
            } elseif ('door' === $deliveryData['deliveryType']) {
                $parameters['courier'] = $deliveryData['courier'];
                foreach ($doorRules as $rule) {
                    $this->sendMessage($order->get_id(), $phone, $rule, $parameters, $contact);
                }
            }
        }
    }

    /**
     * @param array $params
     */
    public function not_picked_up(array $params)
    {
        global $wpdb;

        if (!$this->isEnabled()) {
            return;
        }

        foreach ($params['orders'] as $tmp) {
            /** @var WC_Order $order */
            $order = $tmp['order'];
            $deliveryData = $tmp['deliveryData'];

            $phone = $order->get_billing_phone();
            if ($this->isPhoneBlacklisted($phone)) {
                continue;
            }

            $parameters = $this->getDefaultParameters($order);
            $parameters['office'] = $deliveryData['office'];
            $parameters['days_not_picked_up'] = (string)$deliveryData['workingDaysPassed'];

            $trackingData = $wpdb->get_row("SELECT tracking_number FROM `".$wpdb->prefix."connectix_tracking` WHERE order_id = '".$order->get_id()."'", ARRAY_A);
            $parameters['tracking_number'] = !empty($trackingData['tracking_number']) ? $trackingData['tracking_number'] : '';

            $contact = new Contact();
            $contact->setFirstName($parameters['first_name'])
                ->setLastName($parameters['last_name']);

            foreach ($params['rules'] as $rule) {
                $this->sendMessage($order->get_id(), $phone, $rule, $parameters, $contact);
            }
        }
    }

    /**
     * @param $price
     * @param null $currency
     * @return string
     */
    private function formatPrice($price, $currency = null): string
    {
        $decimalSeparator = wc_get_price_decimal_separator();
        $thousandSeparator = wc_get_price_thousand_separator();
        $decimals = wc_get_price_decimals();
        $price_format = get_woocommerce_price_format();
        $negative = $price < 0;
        $price = apply_filters('raw_woocommerce_price', floatval($negative ? $price * -1 : $price));
        $price = apply_filters('formatted_woocommerce_price', number_format($price, $decimals, $decimalSeparator, $thousandSeparator), $price, $decimals, $decimalSeparator, $thousandSeparator);

        if (apply_filters('woocommerce_price_trim_zeros', false) && $decimals > 0) {
            $price = wc_trim_zeros($price);
        }

        return ($negative ? '-' : '').sprintf($price_format, html_entity_decode(get_woocommerce_currency_symbol($currency)), $price);
    }

    /**
     * @param string $phone
     * @return bool
     */
    private function isPhoneBlacklisted(string $phone): bool
    {
        $blacklist = new Blacklist();
        $blacklist->setPhone($phone);
        try {
            $isBlacklisted = $this->getClient()->getBlacklist()->isBlacklisted($blacklist);
        } catch (Exception\BadConnectionException $e) {
            return false;
        } catch (Exception\BadRequestException $e) {
            return false;
        } catch (Exception\BadResponseBodyException $e) {
            return false;
        }

        return 'yes' === $isBlacklisted;
    }

    /**
     * @param $orderId
     * @param array $data
     */
    public function writeMessage($orderId, array $data)
    {
        global $wpdb;

        $params = [
            'order_id' => $orderId,
            'channel' => $data['channel'],
            'status' => $data['status'],
            'text' => $data['message']['text'],
            'created_at' => $data['createdAt'],
            'provider_id' => $data['id'],
            'type' => $data['type'],
        ];

        if (isset($data['fallback_id'])) {
            $params['fallback_id'] = $data['fallback_id'];
        }

        if ($data['type'] === self::TYPE_READY_FOR_DELIVERY) {
            $params['delivery_type'] = $data['delivery_type'];
        }

        $wpdb->insert($wpdb->prefix.'connectix_messages', $params);
    }

    /**
     * @param $orderId
     * @param $phone
     * @param array $rule
     * @param array $parameters
     * @param Contact|null $contact
     */
    private function sendMessage($orderId, $phone, array $rule, array $parameters = [], Contact $contact = null)
    {
        $enabled = $this->isEnabled();
        $settings = Settings::get_settings();

        if (empty($phone) || !$enabled || empty($settings['token'])) {
            return;
        }

        $queryData = [];
        $integrationToken = $settings['integrationToken'];
        if (!empty($integrationToken)) {
            $queryData = [
                'integration' => $integrationToken,
            ];
        }
        $client = $this->getClient();

        $sandboxUrl = $settings['sandboxUrl'];
        $isSandbox = $settings['system'] === self::SYSTEM_SANDBOX;

        if ($rule['fallback']) {
            $fallback = new Fallback();
            $fallback->setPhone($phone);
            if ($contact && $settings['addContacts']) {
                $fallback->setContact($contact);
            }
            foreach ($rule['templates'] as $tpl) {
                $message = new Message();
                $message->setTemplate($tpl['provider_id'])
                    ->setParameters($parameters);
                if ($isSandbox) {
                    $message->setCallbackUrl($sandboxUrl)
                        ->setInboundUrl($sandboxUrl);
                }

                $fallback->addMessage($message);
            }

            try {
                $response = $client->getFallback()->send($fallback, $queryData);
            } catch (\Exception $e) {
                return;
            }

            foreach ($response['flow'] as $item) {
                $item['fallback_id'] = $response['id'];
                $item['type'] = $rule['type'];
                if ($item['type'] === self::TYPE_READY_FOR_DELIVERY) {
                    $item['delivery_type'] = $rule['delivery_type'];
                }
                $this->writeMessage($orderId, $item);
            }
        } else {
            $request = $client->getMessages();
            foreach ($rule['templates'] as $tpl) {
                $message = new Message();
                $message->setTemplate($tpl['provider_id'])
                    ->setPhone($phone)
                    ->setParameters($parameters);
                if ($isSandbox) {
                    $message->setCallbackUrl($sandboxUrl)
                        ->setInboundUrl($sandboxUrl);
                }

                if ($contact && $settings['addContacts']) {
                    $message->setContact($contact);
                }

                try {
                    $response = $request->send($message, $queryData);
                } catch (\Exception $e) {
                    return;
                }

                $response['type'] = $rule['type'];
                if ($response['type'] === self::TYPE_READY_FOR_DELIVERY) {
                    $response['delivery_type'] = $rule['delivery_type'];
                }
                $this->writeMessage($orderId, $response);
            }
        }
    }
}
